from typing import List, Optional
import tiktoken
from langchain.text_splitter import RecursiveCharacterTextSplitter
import pinecone
import openai
from tqdm.auto import tqdm
import os
from pod_gpt.models import Record, VideoRecord, Metadata

tokenizer = tiktoken.get_encoding('cl100k_base')

# create the length function
def tiktoken_len(text: str) -> int:
    tokens = tokenizer.encode(
        text,
        disallowed_special=()
    )
    return len(tokens)


class Chunker:
    def __init__(self, chunk_size: Optional[int] = 400, chunk_overlap: Optional[int] = 20):
        self.text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,  # number of tokens overlap between chunks
            length_function=tiktoken_len,
            separators=['\n\n', '.\n', '\n', '.', '?', '!', ' ', '']
        )

    def __call__(self, video_record: VideoRecord) -> List[Record]:
        text_chunks = self.text_splitter.split_text(video_record.transcript)
        return [
            Record(
                id=f'{video_record.video_id}-{i}',
                text=text,
                metadata=Metadata(
                    title=video_record.title,
                    channel_id=video_record.channel_id,
                    published=video_record.published,
                    source=video_record.source,
                    chunk=i
                )
            )
            for i, text in enumerate(text_chunks)
        ]


class Indexer:
    dimension_map = {
        'text-embedding-ada-002': 1536
    }
    def __init__(
        self, openai_api_key: Optional[str], pinecone_api_key: Optional[str],
        pinecone_environment: Optional[str], index_name: Optional[str] = "pod-gpt",
        embedding_model_name: Optional[str] = "text-embedding-ada-002",
        chunk_size: Optional[int] = 400, chunk_overlap: Optional[int] = 20
    ):
        # get variables
        self.openai_api_key = openai_api_key or os.environ.get('OPENAI_API_KEY')
        self.pinecone_api_key = pinecone_api_key or os.environ.get('PINECONE_API_KEY')
        self.pinecone_environment = pinecone_environment or os.environ.get('PINECONE_ENVIRONMENT')
        if self.openai_api_key is None:
            raise ValueError('openai_api_key not specified')
        if self.pinecone_api_key is None:
            raise ValueError('pinecone_api_key not specified')
        if self.pinecone_environment is None:
            raise ValueError('pinecone_environment not specified')
        
        # initialize everything else...
        self.chunker = Chunker(chunk_size, chunk_overlap)
        self.embedding_model_name = embedding_model_name
        # initialize pinecone connection
        pinecone.init(
            api_key=pinecone_api_key, environment=pinecone_environment
        )
        # initialize openai connection
        openai.api_key = openai_api_key
        # initialize pinecone index
        if index_name not in pinecone.list_indexes():
            pinecone.create_index(
                index_name, dimension=self.dimension_map[embedding_model_name]
            )
        # connect to Pinecone index
        self.index = pinecone.GRPCIndex(index_name)

    def _embed(self, texts: List[str]) -> List[List[float]]:
        res = openai.Embedding.create(
            input=texts,
            engine=self.embedding_model_name
        )
        return [result["embedding"] for result in res["data"]]

    def _index(self, records: List[Record]) -> None:
        ids = [record.id for record in records]
        texts = [record.text for record in records]
        metadatas = [dict(record.metadata) for record in records]
        # add texts to metadata
        for i, metadata in enumerate(metadatas):
            metadata['text'] = texts[i]
        embeddings = self._embed(texts)
        # upsert to Pinecone index
        self.index.upsert(vectors=zip(ids, embeddings, metadatas))
    
    def __call__(self, video_record: VideoRecord, batch_size: Optional[int] = 100) -> None:
        chunks = self.chunker(video_record)
        for i in range(0, len(chunks), batch_size):
            i_end = min(i + batch_size, len(chunks))
            self._index(chunks[i:i_end])



import os
from pytube import YouTube
from pytube.exceptions import RegexMatchError
from tqdm.auto import tqdm
import requests
import json
from typing import Optional, List, Any
from pod_gpt.models import VideoRecord


class Video:
    def __init__(self, video: VideoRecord):
        self.metadata = video
        try:
            self.yt = YouTube(self.metadata.source)
        except RegexMatchError:
            print(f"RegexMatchError for '{self.metadata.source}'. Video object cannot be created.")

    def _download_mp3(self, save_dir: Optional[str] = './temp'):
        itag = None
        # filters out all the files with "mp4" extension
        files = self.yt.streams.filter(only_audio=True)
        for file in files:
            if file.mime_type == 'audio/mp4':
                itag = file.itag
                break
        if itag is None:
            print("No MP3 audio found.")
        # initialize mp3 file stream
        stream = self.yt.streams.get_by_itag(itag)
        # download the audio
        stream.download(output_path=save_dir, filename=f"{self.metadata.video_id}.mp3")
        self.temp_file = os.path.join(save_dir, f"{self.metadata.video_id}.mp3")
        return self.temp_file

    def _transcribe(self, model: Any):
        # now transcribe
        result = model.transcribe(self.temp_file)
        segments = result['segments']
        # join all segments to get single full transcript for video
        text = ''.join([x['text'] for x in segments])
        self.metadata.transcript = text
        # return
        return text
    
    def transcribe_video(self, model: Any):
        filepath = self._download_mp3()
        transcription = self._transcribe(model)
        # delete mp3 file
        os.unlink(filepath)
        return transcription
    
    def __str__(self):
        return f"{dict(self.metadata)}"


class Channel:
    videos: List[VideoRecord] = []

    def __init__(self, channel_id: str, api_key: str):
        self.channel_id = channel_id
        self.api_key = api_key

    def get_videos_info(self, max_results: Optional[int] = None):
        """Method to scrape all videos and their metadata from a channel

        :param max_results: Maximum number of videos to scrape. If None, all videos are scraped.
        """
        params = {
            "key": self.api_key,
            "part": "snippet",
            "channelId": self.channel_id,
            "type": "video",
            "maxResults": 50
        }
        while True:
            hit_limit = False
            print('.', end='')
            # get a page of video results for this channel
            res = requests.get(
                "https://www.googleapis.com/youtube/v3/search",
                params=params
            )
            # loop through and append the video info to the list
            for record in res.json()['items']:
                _id = record['id']['videoId']
                self.videos.append(VideoRecord(
                    video_id=_id,
                    channel_id=record['snippet']['channelId'],
                    title = record['snippet']['title'],
                    published = record['snippet']['publishedAt'],
                    source = f"https://youtu.be/{_id}",
                    transcript = None
                ))
                # check if we have reached limit
                if max_results is not None and len(self.videos) >= max_results:
                    hit_limit = True
                    break
            if 'nextPageToken' not in res.json().keys() or hit_limit:
                # we have reached the end
                break
            # otherwise we move on to next page
            next_page_token = res.json()['nextPageToken']
            params['pageToken'] = next_page_token
        return {"num_videos": len(self.videos)}
    
    def transcribe_videos(self, model: Any):
        """Method to transcribe all videos in a channel
        """
        for i, video_metadata in enumerate(tqdm(self.videos)):
            video = Video(video_metadata)
            transcription = video.transcribe_video(model)
            self.videos[i].transcript = transcription

    def __repr__(self):
        return f"Channel({self.channel_id})"

    def __str__(self):
        return f"Channel({self.channel_id})"

    def get_videos(self):
        return self.videos

    def save(self, filepath: Optional[str] = 'transcripts.jsonl'):
        # save data
        with open(filepath, 'w') as fp:
            for video in self.videos:
                fp.write(json.dumps(dict(video))+'\n')



from pydantic import BaseModel
from typing import Optional


class Metadata(BaseModel):
    title: str
    channel_id: str
    published: str
    source: str
    chunk: int

class Record(BaseModel):
    id: str
    text: str
    metadata: Metadata

class VideoRecord(BaseModel):
    video_id: str
    channel_id: str
    title: str
    published: str
    source: str
    transcript: Optional[str] = None


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pod_gpt', 'pod_gpt.models']

package_data = \
{'': ['*']}

install_requires = \
['ffmpeg-python>=0.2.0,<0.3.0',
 'langchain>=0.0.125,<0.0.126',
 'openai>=0.27.2,<0.28.0',
 'pinecone-client[grpc]>=2.2.1,<3.0.0',
 'pytube>=12.1.3,<13.0.0',
 'tiktoken>=0.3.3,<0.4.0',
 'torch>=2.0.0,<3.0.0',
 'tqdm>=4.65.0,<5.0.0']

setup_kwargs = {
    'name': 'pod-gpt',
    'version': '0.0.6',
    'description': 'Package for building podcast search indexes.',
    'long_description': '# PodGPT\n\nInstall with:\n\n```\n!pip install pod-gpt\n```\n',
    'author': 'James Briggs',
    'author_email': 'james@aurelio.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
